"""Genius negotiator implementations - TU Delft and misc agents."""

from __future__ import annotations

from ..negotiator import GeniusNegotiator

__all__ = [
    "AI2014Group2",
    "Group1",
    "Group10",
    "Group11",
    "Group12",
    "Group13",
    "Group14",
    "Group15",
    "Group16",
    "Group17",
    "Group18",
    "Group19",
    "Group20",
    "Group21",
    "Group22",
    "Group3Q2015",
    "Group4",
    "Group5",
    "Group6",
    "Group7",
    "Group8",
    "Group9",
    "Q12015Group2",
]


class AI2014Group2(GeniusNegotiator):
    """AI2014Group2 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "agents.ai2014.group2.Group2"
        super().__init__(**kwargs)


class Group1(GeniusNegotiator):
    """Group1 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group1.Group1"
        super().__init__(**kwargs)


class Group10(GeniusNegotiator):
    """Group10 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group10.Group10"
        super().__init__(**kwargs)


class Group11(GeniusNegotiator):
    """Group11 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group11.Group11"
        super().__init__(**kwargs)


class Group12(GeniusNegotiator):
    """Group12 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group12.Group12"
        super().__init__(**kwargs)


class Group13(GeniusNegotiator):
    """Group13 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group13.Group13"
        super().__init__(**kwargs)


class Group14(GeniusNegotiator):
    """Group14 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group14.Group14"
        super().__init__(**kwargs)


class Group15(GeniusNegotiator):
    """Group15 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group15.Group15"
        super().__init__(**kwargs)


class Group16(GeniusNegotiator):
    """Group16 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group16.Group16"
        super().__init__(**kwargs)


class Group17(GeniusNegotiator):
    """Group17 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group17.Group17"
        super().__init__(**kwargs)


class Group18(GeniusNegotiator):
    """Group18 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group18.Group18"
        super().__init__(**kwargs)


class Group19(GeniusNegotiator):
    """Group19 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group19.Group19"
        super().__init__(**kwargs)


class Group20(GeniusNegotiator):
    """Group20 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group20.Group20"
        super().__init__(**kwargs)


class Group21(GeniusNegotiator):
    """Group21 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group21.Group21"
        super().__init__(**kwargs)


class Group22(GeniusNegotiator):
    """Group22 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group22.Group22"
        super().__init__(**kwargs)


class Group3Q2015(GeniusNegotiator):
    """Group3Q2015 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group3.Group3"
        super().__init__(**kwargs)


class Group4(GeniusNegotiator):
    """Group4 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group4.Group4"
        super().__init__(**kwargs)


class Group5(GeniusNegotiator):
    """Group5 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group5.Group5"
        super().__init__(**kwargs)


class Group6(GeniusNegotiator):
    """Group6 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group6.Group6"
        super().__init__(**kwargs)


class Group7(GeniusNegotiator):
    """Group7 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group7.Group7"
        super().__init__(**kwargs)


class Group8(GeniusNegotiator):
    """Group8 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group8.Group8"
        super().__init__(**kwargs)


class Group9(GeniusNegotiator):
    """Group9 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group9.Group9"
        super().__init__(**kwargs)


class Q12015Group2(GeniusNegotiator):
    """Q12015Group2 implementation."""

    def __init__(self, **kwargs):
        """Initialize the instance.

        Args:
            **kwargs: Additional keyword arguments.
        """
        kwargs["java_class_name"] = "parties.in4010.q12015.group2.Group2"
        super().__init__(**kwargs)
