# ruff: noqa: F403
"""Concession strategies for negotiation."""

import negmas.gb.components.concession as _n
from negmas.gb.components.concession import *

__all__ = [_ for _ in _n.__all__ if not _.startswith("GB")]
